//---------------------------------------------------------------------------
#include <vcl.h>
#pragma hdrstop

#include "FarDialog.h"
#include <Common.h>
#include <map>
//---------------------------------------------------------------------------
#pragma package(smart_init)
//---------------------------------------------------------------------------
AnsiString __fastcall StripHotKey(AnsiString Text)
{
  int Len = Text.Length();
  int Pos = 1;
  while (Pos <= Len)
  {
    if (Text[Pos] == '&')
    {
      Text.Delete(Pos, 1);
      Len--;
    }
    else
    {
      Pos++;
    }
  }
  return Text;
}
//---------------------------------------------------------------------------
__fastcall TFarDialog::TFarDialog(TCustomFarPlugin * AFarPlugin) :
  TObject(), FBounds(-1, -1, 40, 10)
{
  assert(AFarPlugin);
  FItems = new TObjectList();
  FFarPlugin = AFarPlugin;
  FFlags = 0;
  FHandle = 0;
  FDefaultGroup = 0;
  FDefaultButton = NULL;
  FNextItemPosition = ipNewLine;
  FDialogItems = NULL;
  FDialogItemsCapacity = 0;
  FRedrawPending = false;
  FChangesPending = false;
  FChangesLocked = 0;
  FResult = -1;

  FBorderBox = new TFarBox(this);
  FBorderBox->Bounds = TRect(3, 1, -4, -2);
  FBorderBox->Double = true;
}
//---------------------------------------------------------------------------
__fastcall TFarDialog::~TFarDialog()
{
  for (int i = 0; i < Items->Count; i++)
  {
    Item[i]->FDialog = NULL;
  }
  delete FItems;
  delete FDialogItems;
  FDialogItemsCapacity = 0;
}
//---------------------------------------------------------------------------
void __fastcall TFarDialog::SetBounds(TRect value)
{
  if (Bounds != value)
  {
    LockChanges();
    try
    {
      FBounds = value;
      if (Handle)
      {
        COORD Coord;
        Coord.X = (short int)Size.x;
        Coord.Y = (short int)Size.y;
        SendMessage(DM_RESIZEDIALOG, 0, (int)&Coord);
        Coord.X = (short int)FBounds.Left;
        Coord.Y = (short int)FBounds.Top;
        SendMessage(DM_MOVEDIALOG, true, (int)&Coord);
      }
      for (int i = 0; i < Items->Count; i++)
      {
        Item[i]->DialogResized();
      }
    }
    __finally
    {
      UnlockChanges();
    }
  }
}
//---------------------------------------------------------------------------
TRect __fastcall TFarDialog::GetClientRect()
{
  TRect R;
  if (FBorderBox)
  {
    R = FBorderBox->Bounds;
    R.Left += 2;
    R.Right -= 2;
    R.Top++;
    R.Bottom--;
  }
  else
  {
    R.Left = 0;
    R.Top = 0;
    R.Bottom = 0;
    R.Right = 0;
  }
  return R;
}
//---------------------------------------------------------------------------
TPoint __fastcall TFarDialog::GetClientSize()
{
  TPoint S;
  if (FBorderBox)
  {
    TRect R = FBorderBox->ActualBounds;
    S.x = R.Width() + 1;
    S.y = R.Height() + 1;
    S.x -= S.x > 4 ? 4 : S.x;
    S.y -= S.y > 2 ? 2 : S.y;
  }
  else
  {
    S = Size;
  }
  return S;
}
//---------------------------------------------------------------------------
void __fastcall TFarDialog::SetHelpTopic(AnsiString value)
{
  if (HelpTopic != value)
  {
    assert(!Handle);
    FHelpTopic = value;
  }
}
//---------------------------------------------------------------------------
void __fastcall TFarDialog::SetFlags(unsigned int value)
{
  if (Flags != value)
  {
    assert(!Handle);
    FFlags = value;
  }
}
//---------------------------------------------------------------------------
void __fastcall TFarDialog::SetCentered(bool value)
{
  if (Centered != value)
  {
    assert(!Handle);
    TRect B = Bounds;
    B.Left = value ? -1 : 0;
    B.Top = value ? -1 : 0;
    Bounds = B;
  }
}
//---------------------------------------------------------------------------
bool __fastcall TFarDialog::GetCentered()
{
  return (Bounds.Left < 0) && (Bounds.Top < 0);
}
//---------------------------------------------------------------------------
TPoint __fastcall TFarDialog::GetSize()
{
  if (Centered)
  {
    return TPoint(Bounds.Right, Bounds.Bottom);
  }
  else
  {
    return TPoint(Bounds.Width() + 1, Bounds.Height() + 1);
  }
}
//---------------------------------------------------------------------------
void __fastcall TFarDialog::SetSize(TPoint value)
{
  TRect B = Bounds;
  if (Centered)
  {
    B.Right = value.x;
    B.Bottom = value.y;
  }
  else
  {
    B.Right = FBounds.Left + value.x;
    B.Bottom = FBounds.Top + value.y;
  }
  Bounds = B;
}
//---------------------------------------------------------------------------
void __fastcall TFarDialog::SetCaption(AnsiString value)
{
  if (Caption != value)
  {
    FBorderBox->Caption = value;
  }
}
//---------------------------------------------------------------------------
AnsiString __fastcall TFarDialog::GetCaption()
{
  return FBorderBox->Caption;
}
//---------------------------------------------------------------------------
int __fastcall TFarDialog::GetItemCount()
{
  return FItems->Count;
}
//---------------------------------------------------------------------------
TFarDialogItem * __fastcall TFarDialog::GetItem(int Index)
{
  TFarDialogItem * DialogItem;
  if (FItems->Count)
  {
    assert(Index >= 0 && Index < FItems->Count);
    DialogItem = dynamic_cast<TFarDialogItem *>(Items->Items[Index]);
    assert(DialogItem);
  }
  else
  {
    DialogItem = NULL;
  }
  return DialogItem;
}
//---------------------------------------------------------------------------
void __fastcall TFarDialog::Add(TFarDialogItem * DialogItem)
{
  TRect R = ClientRect;
  int Left, Top;
  GetNextItemPosition(Left, Top);
  R.Left = Left;
  R.Top = Top;

  if (FDialogItemsCapacity == Items->Count)
  {
    int DialogItemsDelta = 10;
    FarDialogItem * NewDialogItems;
    NewDialogItems = new FarDialogItem[Items->Count + DialogItemsDelta];
    if (FDialogItems)
    {
      memcpy(NewDialogItems, FDialogItems, FDialogItemsCapacity * sizeof(FarDialogItem));
      delete FDialogItems;
    }
    memset(NewDialogItems + FDialogItemsCapacity, 0,
      DialogItemsDelta * sizeof(FarDialogItem));
    FDialogItems = NewDialogItems;
    FDialogItemsCapacity += DialogItemsDelta;
  }

  assert(DialogItem);
  DialogItem->FItem = Items->Add(DialogItem);

  R.Bottom = R.Top;
  DialogItem->Bounds = R;
  DialogItem->Group = DefaultGroup;
}
//---------------------------------------------------------------------------
void __fastcall TFarDialog::GetNextItemPosition(int & Left, int & Top)
{
  TRect R = ClientRect;
  Left = R.Left;
  Top = R.Top;
  
  TFarDialogItem * LastI = Item[Items->Count-1];
  LastI = LastI == FBorderBox ? NULL : LastI;

  if (LastI)
  {
    switch (NextItemPosition) {
      case ipNewLine:
        Top = LastI->Bottom + 1;
        break;

      case ipBelow:
        Top = LastI->Bottom + 1;
        Left = LastI->Left;
        break;

      case ipRight:
        Top = LastI->Top;
        Left = LastI->Right + 3;
        break;
    }
  }
}
//---------------------------------------------------------------------------
long WINAPI TFarDialog::DialogProcGeneral(HANDLE Handle, int Msg, int Param1, long Param2)
{
  static std::map<HANDLE, long> Dialogs;
  TFarDialog * Dialog;
  if (Msg == DN_INITDIALOG)
  {
    assert(Dialogs.find(Handle) == Dialogs.end());
    Dialogs[Handle] = Param2;
    Dialog = (TFarDialog *)Param2;
    Dialog->FHandle = Handle;
  }
  else
  {
    assert(Dialogs.find(Handle) != Dialogs.end());
    Dialog = (TFarDialog *)Dialogs[Handle];
  }
  long Result = Dialog->DialogProc(Msg, Param1, Param2);

  if ((Msg == DN_CLOSE) && Result)
  {
    Dialog->FHandle = 0;
    Dialogs.erase(Handle);
  }
  return Result;
}
//---------------------------------------------------------------------------
long __fastcall TFarDialog::DialogProc(int Msg, int Param1, long Param2)
{
  long Result = 0;
  bool Handled = false;

  try
  {
    bool Changed = false;

    switch (Msg) {
      case DN_BTNCLICK:
      case DN_EDITCHANGE:
      case DN_GOTFOCUS:
      case DN_KILLFOCUS:
      case DN_LISTCHANGE:
        Changed = true;

      case DN_MOUSECLICK:
      case DN_CTLCOLORDLGITEM:
      case DN_DRAWDLGITEM:
      case DN_HOTKEY:
      case DN_KEY:
        if (Param1 >= 0)
        {
          TFarDialogItem * I = Item[Param1];
          try
          {
            Result = I->ItemProc(Msg, Param2);
          }
          catch(Exception & E)
          {
            Handled = true;
            FarPlugin->HandleException(&E);
            Result = I->FailItemProc(Msg, Param2);
          }

          if (!Result && (Msg == DN_KEY) && OnKey)
          {
            bool BResult;
            BResult = false;
            OnKey(this, I, Param2, BResult);
            Result = BResult;
          }
          Handled = true;
        }
        break;
    }
    if (!Handled)
    {
      switch (Msg) {
        case DN_INITDIALOG:
          Init();
          Result = true;
          break;

        case DN_DRAGGED:
          if (Param1 == 1)
          {
            RefreshBounds();
          }
          break;

        case DN_CLOSE:
          if (Param1 >= 0)
          {
            TFarButton * Button = dynamic_cast<TFarButton *>(Item[Param1]);
            assert(Button);
            FResult = Button->Result;
          }
          else
          {
            FResult = -1;
          }
          Result = CloseQuery();
          if (!Result)
          {
            FResult = -1; 
          }
          Handled = true;
          break;

        case DN_ENTERIDLE:
          Idle();
          break;
      };
    
      if (!Handled)
      {
        Result = DefaultDialogProc(Msg, Param1, Param2);
      }
    }
    if (Changed)
    {
      Change();
    }
  }
  catch(Exception & E)
  {
    FarPlugin->HandleException(&E);
    if (!Handled)
    {
      Result = FailDialogProc(Msg, Param1, Param2);
    }
  }
  return Result;
}
//---------------------------------------------------------------------------
long __fastcall TFarDialog::DefaultDialogProc(int Msg, int Param1, long Param2)
{
  return FarPlugin->FStartupInfo.DefDlgProc(Handle, Msg, Param1, Param2);
}
//---------------------------------------------------------------------------
long __fastcall TFarDialog::FailDialogProc(int Msg, int Param1, long Param2)
{
  long Result;
  switch (Msg) {
    case DN_CLOSE:
      Result = false;
      break;

    default:
      Result = DefaultDialogProc(Msg, Param1, Param2);
      break;
  }
  return Result;
}
//---------------------------------------------------------------------------
void __fastcall TFarDialog::Idle()
{
  // nothing
}
//---------------------------------------------------------------------------
bool __fastcall TFarDialog::CloseQuery()
{
  bool Result = true;
  for (int i = 0; i < Items->Count && Result; i++)
  {
    if (!Item[i]->CloseQuery())
    {
      Result = false;
    }
  }
  return Result;
}
//---------------------------------------------------------------------------
void __fastcall TFarDialog::RefreshBounds()
{
  SMALL_RECT Rect;
  SendMessage(DM_GETDLGRECT, 0, (int)&Rect);
  FBounds.Left = Rect.Left;
  FBounds.Top = Rect.Top;
  FBounds.Right = Rect.Right;
  FBounds.Bottom = Rect.Bottom;
}
//---------------------------------------------------------------------------
void __fastcall TFarDialog::Init()
{
  for (int i = 0; i < Items->Count; i++)
  {
    Item[i]->Init();
  }

  RefreshBounds();

  Change();
}
//---------------------------------------------------------------------------
int __fastcall TFarDialog::ShowModal()
{
  FResult = -1;
  
  int BResult;

  TFarDialog * PrevTopDialog = FarPlugin->FTopDialog;
  FarPlugin->FTopDialog = this;
  try
  {
    assert(DefaultButton);
    assert(DefaultButton->Default);

    BResult = FarPlugin->FStartupInfo.DialogEx(FarPlugin->FStartupInfo.ModuleNumber,
      Bounds.Left, Bounds.Top, Bounds.Right, Bounds.Bottom,
      (char *)HelpTopic.data(), FDialogItems, Items->Count, 0, Flags,
      DialogProcGeneral, (long)this);

    if (BResult >= 0)
    {
      TFarButton * Button = dynamic_cast<TFarButton *>(Item[BResult]);
      assert(Button);
      // correct result should be already set by TFarButton
      assert(FResult == Button->Result);
      FResult = Button->Result;
    }
    else
    {
      // allow only one negative value = -1
      FResult = -1;
    }
  }
  __finally
  {
    FarPlugin->FTopDialog = PrevTopDialog;
  }

  return FResult;
}
//---------------------------------------------------------------------------
void __fastcall TFarDialog::Change()
{
  if (FChangesLocked > 0)
  {
    FChangesPending = true;
  }
  else
  {
    TList * NotifiedContainers = new TList();
    try
    {
      TFarDialogItem * DItem;
      for (int i = 0; i < Items->Count; i++)
      {
        DItem = Item[i];
        DItem->Change();
        if (DItem->Container && NotifiedContainers->IndexOf(DItem->Container) < 0)
        {
          NotifiedContainers->Add(DItem->Container);
        }
      }

      for (int Index = 0; Index < NotifiedContainers->Count; Index++)
      {
        ((TFarDialogContainer *)NotifiedContainers->Items[Index])->Change();
      }
    }
    __finally
    {
      delete NotifiedContainers;
    }
  }
}
//---------------------------------------------------------------------------
long __fastcall TFarDialog::SendMessage(int Msg, int Param1, int Param2)
{
  assert(Handle);
  return FarPlugin->FStartupInfo.SendDlgMessage(Handle, Msg, Param1, Param2);
}
//---------------------------------------------------------------------------
void __fastcall TFarDialog::Redraw()
{
  SendMessage(DM_REDRAW, 0, 0);
  FRedrawPending = false;
}
//---------------------------------------------------------------------------
void __fastcall TFarDialog::ShowGroup(int Group, bool Show)
{
  LockChanges();
  try
  {
    for (int i = 0; i < Items->Count; i++)
    {
      TFarDialogItem * I = Item[i];
      if (I->Group == Group)
      {
        I->Visible = Show;
      }
    }
  }
  __finally
  {
    UnlockChanges();
  }
}
//---------------------------------------------------------------------------
void __fastcall TFarDialog::SetItemFocused(TFarDialogItem * value)
{
  if (value != ItemFocused)
  {
    assert(value);
    value->SetFocus();
  }
}
//---------------------------------------------------------------------------
AnsiString __fastcall TFarDialog::GetMsg(int MsgId)
{
  return FFarPlugin->GetMsg(MsgId);
}
//---------------------------------------------------------------------------
void __fastcall TFarDialog::LockChanges()
{
  assert(FChangesLocked < 10);
  FChangesLocked++;
  if (FChangesLocked == 1)
  {
    assert(!FChangesPending);
    if (Handle)
    {
      SendMessage(DM_ENABLEREDRAW, false, 0);
    }
  }
}
//---------------------------------------------------------------------------
void __fastcall TFarDialog::UnlockChanges()
{
  assert(FChangesLocked > 0);
  FChangesLocked--;
  if (FChangesLocked == 0)
  {
    try
    {
      if (FChangesPending)
      {
        FChangesPending = false;
        Change();
      }
    }
    __finally
    {
      if (Handle)
      {
        SendMessage(DM_ENABLEREDRAW, true, 0);
      }
    }
  }
}
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
__fastcall TFarDialogContainer::TFarDialogContainer(TFarDialog * ADialog) :
  TObject()
{
  assert(ADialog);
  
  FItems = new TObjectList();
  FItems->OwnsObjects = false;
  FDialog = ADialog;
  FEnabled = true;

  Dialog->GetNextItemPosition(FLeft, FTop);
}
//---------------------------------------------------------------------------
__fastcall TFarDialogContainer::~TFarDialogContainer()
{
  delete FItems;
}
//---------------------------------------------------------------------------
AnsiString __fastcall TFarDialogContainer::GetMsg(int MsgId)
{
  return Dialog->GetMsg(MsgId);
}
//---------------------------------------------------------------------------
void __fastcall TFarDialogContainer::Add(TFarDialogItem * Item)
{
  assert(FItems->IndexOf(Item) < 0);
  Item->Container = this;
  FItems->Add(Item);
}
//---------------------------------------------------------------------------
void __fastcall TFarDialogContainer::Remove(TFarDialogItem * Item)
{
  assert(FItems->IndexOf(Item) >= 0);
  Item->Container = NULL;
  FItems->Remove(Item);
  if (FItems->Count == 0)
  {
    delete this;
  }
}
//---------------------------------------------------------------------------
void __fastcall TFarDialogContainer::SetPosition(int Index, int value)
{
  int & Position = Index ? FTop : FLeft;
  if (Position != value)
  {
    Position = value;
    for (int Index = 0; Index < FItems->Count; Index++)
    {
      dynamic_cast<TFarDialogItem *>(FItems->Items[Index])->DialogResized();
    }
  }
}
//---------------------------------------------------------------------------
void __fastcall TFarDialogContainer::Change()
{
}
//---------------------------------------------------------------------------
void __fastcall TFarDialogContainer::SetEnabled(bool value)
{
  if (Enabled != value)
  {
    FEnabled = true;
    for (int Index = 0; Index < FItems->Count; Index++)
    {
      dynamic_cast<TFarDialogItem *>(FItems->Items[Index])->UpdateEnabled();
    }
  }
}
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
__fastcall TFarDialogItem::TFarDialogItem(TFarDialog * ADialog, int AType) :
  TObject()
{
  assert(ADialog);
  FDialog = ADialog;
  FDefaultType = AType;
  FTag = 0;
  FGroup = 0;
  FEnabled = true;
  FIsEnabled = true;
  FOnExit = NULL;
  FInternalDialogItem = NULL;
  FColors = 0;
  FColorMask = 0;

  Dialog->Add(this);

  DialogItem->Type = AType;
}
//---------------------------------------------------------------------------
__fastcall TFarDialogItem::~TFarDialogItem()
{
  assert(!Dialog);
}
//---------------------------------------------------------------------------
FarDialogItem * __fastcall TFarDialogItem::GetDialogItem()
{
  assert(Dialog);
  return &Dialog->FDialogItems[Item];
}
//---------------------------------------------------------------------------
void __fastcall TFarDialogItem::SetBounds(TRect value)
{
  if (Bounds != value)
  {
    FBounds = value;
    UpdateBounds();
  }
}
//---------------------------------------------------------------------------
void __fastcall TFarDialogItem::DialogResized()
{
  UpdateBounds();
}
//---------------------------------------------------------------------------
void __fastcall TFarDialogItem::UpdateBounds()
{
  TRect B = Bounds;
  FarDialogItem * DItem = DialogItem;
  #define BOUND(DIB, BB, DB, CB) DItem->DIB = B.BB >= 0 ? \
    (Container ? Container->CB : 0 ) + B.BB : Dialog->Size.DB + B.BB
  BOUND(X1, Left, x, Left);
  BOUND(Y1, Top, y, Top);
  BOUND(X2, Right, x, Left);
  BOUND(Y2, Bottom, y, Top);
  #undef BOUND

  if (Dialog->Handle)
  {
    B = ActualBounds;
    SMALL_RECT Rect;
    Rect.Left = (short int)B.Left;
    Rect.Top = (short int)B.Top;
    Rect.Right = (short int)B.Right;
    Rect.Bottom = (short int)B.Bottom;
    SendMessage(DM_SETITEMPOSITION, (int)&Rect);
  }
}
//---------------------------------------------------------------------------
char __fastcall TFarDialogItem::GetColor(int Index)
{
  return *(((char *)&FColors) + Index);
}
//---------------------------------------------------------------------------
void __fastcall TFarDialogItem::SetColor(int Index, char value)
{
  if (GetColor(Index) != value)
  {
    *(((char *)&FColors) + Index) = value;
    FColorMask |= (0xFF << (Index * 8));
  }
}
//---------------------------------------------------------------------------
void __fastcall TFarDialogItem::SetFlags(unsigned int value)
{
  if (Flags != value)
  {
    assert(!Dialog->Handle);
    UpdateFlags(value);
  }
}
//---------------------------------------------------------------------------
void __fastcall TFarDialogItem::UpdateFlags(unsigned int value)
{
  if (Flags != value)
  {
    DialogItem->Flags = value;
    DialogChange();
  }
}
//---------------------------------------------------------------------------
TRect __fastcall TFarDialogItem::GetActualBounds()
{
  return TRect(DialogItem->X1, DialogItem->Y1,
    DialogItem->X2, DialogItem->Y2);
}
//---------------------------------------------------------------------------
unsigned int __fastcall TFarDialogItem::GetFlags()
{
  return DialogItem->Flags;
}
//---------------------------------------------------------------------------
void __fastcall TFarDialogItem::SetData(const AnsiString value)
{
  if (Data != value)
  {
    AnsiString LimitedStr = value.SubString(1, sizeof(DialogItem->Data) - 1);
    if (Dialog->Handle)
    {
      SendMessage(DM_SETTEXTPTR, (int)LimitedStr.c_str());
    }
    strcpy(DialogItem->Data, LimitedStr.c_str());
    DialogChange();
  }
}
//---------------------------------------------------------------------------
void __fastcall TFarDialogItem::UpdateData(const AnsiString value)
{
  strcpy(DialogItem->Data, value.SubString(1, sizeof(DialogItem->Data) - 1).c_str());
}
//---------------------------------------------------------------------------
AnsiString __fastcall TFarDialogItem::GetData()
{
  return DialogItem->Data;
}
//---------------------------------------------------------------------------
void __fastcall TFarDialogItem::SetType(int value)
{
  if (Type != value)
  {
    assert(!Dialog->Handle);
    DialogItem->Type = value;
  }
}
//---------------------------------------------------------------------------
int __fastcall TFarDialogItem::GetType()
{
  return DialogItem->Type;
}
//---------------------------------------------------------------------------
void __fastcall TFarDialogItem::SetAlterType(int Index, bool value)
{
  if (GetAlterType(Index) != value)
  {
    Type = value ? Index : FDefaultType;
  }
}
//---------------------------------------------------------------------------
bool __fastcall TFarDialogItem::GetAlterType(int Index)
{
  return (Type == Index);
}
//---------------------------------------------------------------------------
bool __fastcall TFarDialogItem::GetFlag(int Index)
{
  bool Result = (Flags & (Index & 0xFFFFFF00UL)) != 0;
  if (Index & 0x000000FFUL)
  {
    Result = !Result;
  }
  return Result;
}
//---------------------------------------------------------------------------
void __fastcall TFarDialogItem::SetFlag(int Index, bool value)
{
  if (GetFlag(Index) != value)
  {
    if (Index & DIF_INVERSE)
    {
      value = !value;
    }

    unsigned long F = Flags;
    unsigned long Flag = Index & 0xFFFFFF00UL;
    bool ToHandle = true;

    switch (Flag) {
      case DIF_DISABLE:
        if (Dialog->Handle)
        {
          SendMessage(DM_ENABLE, !value);
        }
        break;

      case DIF_HIDDEN:
        if (Dialog->Handle)
        {
          SendMessage(DM_SHOWITEM, !value);
        }
        break;
    }

    if (ToHandle)
    {
      if (value)
      {
        F |= Flag;
      }
      else
      {
        F &= ~Flag;
      }
      UpdateFlags(F);
    }
  }
}
//---------------------------------------------------------------------------
void __fastcall TFarDialogItem::SetEnabledDependency(TFarDialogItem * value)
{
  if (EnabledDependency != value)
  {
    FEnabledDependency = value;
    Change();
  }
}
//---------------------------------------------------------------------------
void __fastcall TFarDialogItem::SetEnabledDependencyNegative(TFarDialogItem * value)
{
  if (EnabledDependencyNegative != value)
  {
    FEnabledDependencyNegative = value;
    Change();
  }
}
//---------------------------------------------------------------------------
bool __fastcall TFarDialogItem::GetIsEmpty()
{
  return Data.Trim().IsEmpty();
}
//---------------------------------------------------------------------------
long __fastcall TFarDialogItem::FailItemProc(int Msg, long Param)
{
  long Result;
  switch (Msg) {
    case DN_KILLFOCUS:
      Result = Item;
      break;

    default:
      Result = DefaultItemProc(Msg, Param);
      break;
  }
  return Result;
}
//---------------------------------------------------------------------------
long __fastcall TFarDialogItem::ItemProc(int Msg, long Param)
{
  if (Msg == DN_DRAWDLGITEM)
  {
    //assert(!FInternalDialogItem || FInternalDialogItem == (FarDialogItem *)Param);
    FInternalDialogItem = (FarDialogItem *)Param;
  }
  else if (Msg == DN_GOTFOCUS)
  {
    DoFocus();
    UpdateFocused(true);
  }
  else if (Msg == DN_KILLFOCUS)
  {
    DoExit();
    UpdateFocused(false);
  }

  long Result;
  Result = DefaultItemProc(Msg, Param);
  if (Msg == DN_CTLCOLORDLGITEM && FColorMask)
  {
    Result &= ~FColorMask;
    Result |= (FColors & FColorMask);
  }
  return Result;
}
//---------------------------------------------------------------------------
void __fastcall TFarDialogItem::DoFocus()
{
}
//---------------------------------------------------------------------------
void __fastcall TFarDialogItem::DoExit()
{
  if (OnExit)
  {
    OnExit(this);
  }
}
//---------------------------------------------------------------------------
long __fastcall TFarDialogItem::DefaultItemProc(int Msg, int Param)
{
  return Dialog->FarPlugin->FStartupInfo.DefDlgProc(Dialog->Handle, Msg, Item, Param);
}
//---------------------------------------------------------------------------
void __fastcall TFarDialogItem::Change()
{
  if (EnabledDependency || EnabledDependencyNegative)
  {
    UpdateEnabled();
  }
}
//---------------------------------------------------------------------------
void __fastcall TFarDialogItem::SetEnabled(bool value)
{
  if (Enabled != value)
  {
    FEnabled = value;
    UpdateEnabled();
  }
}
//---------------------------------------------------------------------------
void __fastcall TFarDialogItem::UpdateEnabled()
{
  bool Value =
    Enabled &&
    (!EnabledDependency ||
      (!EnabledDependency->IsEmpty && EnabledDependency->IsEnabled)) &&
    (!EnabledDependencyNegative ||
      (EnabledDependencyNegative->IsEmpty || !EnabledDependencyNegative->IsEnabled)) &&
    (!Container || Container->Enabled);

  if (Value != IsEnabled)
  {
    FIsEnabled = Value;
    SetFlag(DIF_DISABLE | DIF_INVERSE, IsEnabled);
  }
}
//---------------------------------------------------------------------------
void __fastcall TFarDialogItem::DialogChange()
{
  assert(Dialog);
  Dialog->Change();
}
//---------------------------------------------------------------------------
long __fastcall TFarDialogItem::SendMessage(int Msg, int Param)
{
  return Dialog->SendMessage(Msg, Item, Param);
}
//---------------------------------------------------------------------------
void __fastcall TFarDialogItem::SetSelected(int value)
{
  if (Selected != value)
  {
    if (Dialog->Handle)
    {
      SendMessage(DM_SETCHECK, value);
    }
    UpdateSelected(value);
  }
}
//---------------------------------------------------------------------------
void __fastcall TFarDialogItem::UpdateSelected(int value)
{
  if (Selected != value)
  {
    DialogItem->Selected = value;
    DialogChange();
  }
}
//---------------------------------------------------------------------------
int __fastcall TFarDialogItem::GetSelected()
{
  return DialogItem->Selected;
}
//---------------------------------------------------------------------------
void __fastcall TFarDialogItem::SetChecked(bool value)
{
  Selected = value ? BSTATE_CHECKED : BSTATE_UNCHECKED;
}
//---------------------------------------------------------------------------
bool _fastcall TFarDialogItem::GetChecked()
{
  return Selected == BSTATE_CHECKED;
}
//---------------------------------------------------------------------------
void __fastcall TFarDialogItem::Move(int DeltaX, int DeltaY)
{
  TRect R = Bounds;

  R.Left += DeltaX;
  R.Right += DeltaX;
  R.Top += DeltaY;
  R.Bottom += DeltaY;

  Bounds = R;
}
//---------------------------------------------------------------------------
void __fastcall TFarDialogItem::SetCoordinate(int Index, int value)
{
  assert(sizeof(TRect) == sizeof(long) * 4);
  TRect R = Bounds;
  long * D = (long *)&R;
  D += Index;
  *D = value;
  Bounds = R;
}
//---------------------------------------------------------------------------
int __fastcall TFarDialogItem::GetCoordinate(int Index)
{
  assert(sizeof(TRect) == sizeof(long) * 4);
  TRect R = Bounds;
  long * D = (long *)&R;
  D += Index;
  return *D;
}
//---------------------------------------------------------------------------
void __fastcall TFarDialogItem::SetWidth(int value)
{
  TRect R = Bounds;
  if (R.Left >= 0)
  {
    R.Right = R.Left + value - 1;
  }
  else
  {
    assert(R.Right < 0);
    R.Left = R.Right - value + 1;
  }
  Bounds = R;
}
//---------------------------------------------------------------------------
int __fastcall TFarDialogItem::GetWidth()
{
  return ActualBounds.Width() + 1;
}
//---------------------------------------------------------------------------
void __fastcall TFarDialogItem::SetHeight(int value)
{
  TRect R = Bounds;
  if (R.Top >= 0)
  {
    R.Bottom = R.Top + value - 1;
  }
  else
  {
    assert(R.Bottom < 0);
    R.Top = R.Bottom - value + 1;
  }
  Bounds = R;
}
//---------------------------------------------------------------------------
int __fastcall TFarDialogItem::GetHeight()
{
  return ActualBounds.Height() + 1;
}
//---------------------------------------------------------------------------
bool __fastcall TFarDialogItem::CanFocus()
{
  return Visible && Enabled &&
    (Type == DI_EDIT || Type == DI_PSWEDIT || Type == DI_FIXEDIT ||
     Type == DI_BUTTON || Type == DI_CHECKBOX	|| Type == DI_RADIOBUTTON	||
     Type == DI_COMBOBOX	|| Type == DI_LISTBOX);
}
//---------------------------------------------------------------------------
bool __fastcall TFarDialogItem::Focused()
{
  return DialogItem->Focus != 0;
}
//---------------------------------------------------------------------------
void __fastcall TFarDialogItem::UpdateFocused(bool value)
{
  DialogItem->Focus = value;
  assert(Dialog);
  Dialog->FItemFocused = value ? this : NULL;
}
//---------------------------------------------------------------------------
void __fastcall TFarDialogItem::SetFocus()
{
  assert(CanFocus());
  if (!Focused())
  {
    if (Dialog->Handle)
    {
      SendMessage(DM_SETFOCUS, 0);
    }
    else
    {
      if (Dialog->ItemFocused)
      {
        assert(Dialog->ItemFocused != this);
        Dialog->ItemFocused->UpdateFocused(false);
      }
      UpdateFocused(true);
    }
  }
}
//---------------------------------------------------------------------------
void __fastcall TFarDialogItem::Init()
{
  if (GetFlag(DIF_CENTERGROUP))
  {
    SMALL_RECT Rect;

    SendMessage(DM_GETITEMPOSITION, (int)&Rect);

    TRect B = Bounds;
    B.Left = Rect.Left;
    B.Right = Rect.Right;
    Bounds = B;
  }
}
//---------------------------------------------------------------------------
bool __fastcall TFarDialogItem::CloseQuery()
{
  if (Focused() && (Dialog->Result >= 0))
  {
    DoExit();
  }
  return true;
}
//---------------------------------------------------------------------------
void __fastcall TFarDialogItem::SetContainer(TFarDialogContainer * value)
{
  if (Container != value)
  {
    TFarDialogContainer * PrevContainer = Container;
    FContainer = value;
    if (PrevContainer)
    {
      PrevContainer->Remove(this);
    }
    if (Container)
    {
      Container->Add(this);
    }
    UpdateBounds();
    UpdateEnabled();
  }
}
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
__fastcall TFarBox::TFarBox(TFarDialog * ADialog) :
  TFarDialogItem(ADialog, DI_SINGLEBOX)
{
}
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
__fastcall TFarButton::TFarButton(TFarDialog * ADialog) :
  TFarDialogItem(ADialog, DI_BUTTON)
{
  FResult = 0;
}
//---------------------------------------------------------------------------
void __fastcall TFarButton::SetData(const AnsiString value)
{
  TFarDialogItem::SetData(value);
  if (Left >= 0 || Right >= 0)
  {
    Width = 2 + StripHotKey(value).Length() + 2;
  }
}
//---------------------------------------------------------------------------
void __fastcall TFarButton::SetDefault(bool value)
{
  if (Default != value)
  {
    assert(!Dialog->Handle);
    DialogItem->DefaultButton = value;
    if (value)
    {
      if (Dialog->DefaultButton && (Dialog->DefaultButton != this))
      {
        Dialog->DefaultButton->Default = false;
      }
      Dialog->FDefaultButton = this;
    }
    else if (Dialog->DefaultButton == this)
    {
      Dialog->FDefaultButton = NULL;
    }
    DialogChange();
  }
}
//---------------------------------------------------------------------------
bool __fastcall TFarButton::GetDefault()
{
  return DialogItem->DefaultButton;
}
//---------------------------------------------------------------------------
long __fastcall TFarButton::ItemProc(int Msg, long Param)
{
  if (Msg == DN_BTNCLICK)
  {
    if (!Enabled)
    {
      return true;
    }
    else
    {
      bool Close = (Result != 0);
      if (OnClick)
      {
        OnClick(this, Close);
      }
      if (!Close)
      {
        return true;
      }
    }
  }
  return TFarDialogItem::ItemProc(Msg, Param);
}
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
__fastcall TFarCheckBox::TFarCheckBox(TFarDialog * ADialog) :
  TFarDialogItem(ADialog, DI_CHECKBOX)
{
}
//---------------------------------------------------------------------------
long __fastcall TFarCheckBox::ItemProc(int Msg, long Param)
{
  if (Msg == DN_BTNCLICK)
  {
    bool Allow = true;
    if (OnAllowChange)
    {
      OnAllowChange(this, Param, Allow);
    }
    if (Allow)
    {
      UpdateSelected(Param);
    }
    return Allow;
  }
  else
  {
    return TFarDialogItem::ItemProc(Msg, Param);
  }
}
//---------------------------------------------------------------------------
bool __fastcall TFarCheckBox::GetIsEmpty()
{
  return Checked != BSTATE_CHECKED;
}
//---------------------------------------------------------------------------
void __fastcall TFarCheckBox::SetData(const AnsiString value)
{
  TFarDialogItem::SetData(value);
  if (Left >= 0 || Right >= 0)
  {
    Width = 4 + StripHotKey(value).Length();
  }
}
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
__fastcall TFarRadioButton::TFarRadioButton(TFarDialog * ADialog) :
  TFarDialogItem(ADialog, DI_RADIOBUTTON)
{
}
//---------------------------------------------------------------------------
long __fastcall TFarRadioButton::ItemProc(int Msg, long Param)
{
  if (Msg == DN_BTNCLICK)
  {
    bool Allow = true;
    if (OnAllowChange)
    {
      OnAllowChange(this, Param, Allow);
    }
    if (Allow)
    {
      // This does not correspond to FAR API Manual, but it works so.
      // Manual says that Param should contain ID of previously selected dialog item
      UpdateSelected(Param);
    }
    return Allow;
  }
  else
  {
    return TFarDialogItem::ItemProc(Msg, Param);
  }
}
//---------------------------------------------------------------------------
bool __fastcall TFarRadioButton::GetIsEmpty()
{
  return !Checked;
}
//---------------------------------------------------------------------------
void __fastcall TFarRadioButton::SetData(const AnsiString value)
{
  TFarDialogItem::SetData(value);
  if (Left >= 0 || Right >= 0)
  {
    Width = 4 + StripHotKey(value).Length();
  }
}
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
__fastcall TFarEdit::TFarEdit(TFarDialog * ADialog) :
  TFarDialogItem(ADialog, DI_EDIT)
{
  AutoSelect = true;
}
//---------------------------------------------------------------------------
long __fastcall TFarEdit::ItemProc(int Msg, long Param)
{
  if (Msg == DN_EDITCHANGE)
  {
    strcpy(DialogItem->Data, ((FarDialogItem *)Param)->Data);
  }
  return TFarDialogItem::ItemProc(Msg, Param);
}
//---------------------------------------------------------------------------
AnsiString __fastcall TFarEdit::GetHistoryMask(int Index)
{
  return
    ((Index == 0) && (Flags & DIF_HISTORY)) ||
    ((Index == 1) && (Flags & DIF_MASKEDIT)) ? DialogItem->Mask : "";
}
//---------------------------------------------------------------------------
void __fastcall TFarEdit::SetHistoryMask(int Index, AnsiString value)
{
  if (GetHistoryMask(Index) != value)
  {
    assert(!Dialog->Handle);
    assert(&DialogItem->Mask == &DialogItem->History);

    delete DialogItem->Mask;
    if (value.IsEmpty())
    {
      DialogItem->Mask = NULL;
    }
    else
    {
      DialogItem->Mask = new char[value.Length()+1];
      strcpy((char *)DialogItem->Mask, value.c_str());
    }
    bool PrevHistory = !History.IsEmpty();
    SetFlag(DIF_HISTORY, (Index == 0) && !value.IsEmpty());
    bool Masked = (Index == 1) && !value.IsEmpty();
    SetFlag(DIF_MASKEDIT, Masked);
    if (Masked)
    {
      Fixed = true;
    }
    bool CurrHistory = !History.IsEmpty();
    if (PrevHistory != CurrHistory)
    {
      // add/remove space for history arrow
      Width = Width + (CurrHistory ? -1 : 1);
    }
    DialogChange();
  }
}
//---------------------------------------------------------------------------
void __fastcall TFarEdit::SetAsInteger(int value)
{
  int Int;
  if (!TryStrToInt(Text.Trim(), Int) || (AsInteger != value))
  {
    Text = IntToStr(value);
    DialogChange();
  }
}
//---------------------------------------------------------------------------
int __fastcall TFarEdit::GetAsInteger()
{
  return StrToIntDef(Text.Trim(), 0);
}
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
__fastcall TFarSeparator::TFarSeparator(TFarDialog * ADialog) :
  TFarDialogItem(ADialog, DI_TEXT)
{
  Left = -1;
  SetFlag(DIF_SEPARATOR, true);
}
//---------------------------------------------------------------------------
void __fastcall TFarSeparator::UpdateBounds()
{
  TFarDialogItem::UpdateBounds();
  if (Bounds.Left < 0)
  {
    DialogItem->X1 = -1;
  }
}
//---------------------------------------------------------------------------
void __fastcall TFarSeparator::SetDouble(bool value)
{
  if (Double != value)
  {
    assert(!Dialog->Handle);
    SetFlag(DIF_SEPARATOR, !value);
    SetFlag(DIF_SEPARATOR2, value);
  }
}
//---------------------------------------------------------------------------
bool __fastcall TFarSeparator::GetDouble()
{
  return GetFlag(DIF_SEPARATOR2);
}
//---------------------------------------------------------------------------
void __fastcall TFarSeparator::SetPosition(int value)
{
  TRect R = Bounds;
  R.Top = value;
  R.Bottom = value;
  Bounds = R;
}
//---------------------------------------------------------------------------
int __fastcall TFarSeparator::GetPosition()
{
  return Bounds.Top;
}
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
__fastcall TFarText::TFarText(TFarDialog * ADialog) :
  TFarDialogItem(ADialog, DI_TEXT)
{
}
//---------------------------------------------------------------------------
void __fastcall TFarText::SetData(const AnsiString value)
{
  TFarDialogItem::SetData(value);
  if (Left >= 0 || Right >= 0)
  {
    Width = StripHotKey(value).Length();
  }
}
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
__fastcall TFarList::TFarList(TFarDialogItem * ADialogItem) :
  TStringList()
{
  assert(ADialogItem);
  assert(ADialogItem->Type == DI_COMBOBOX || ADialogItem->Type == DI_LISTBOX);

  FDialogItem = ADialogItem;
  FListItems = new FarList;
  memset(FListItems, 0, sizeof(*FListItems));
  FNoDialogUpdate = false;
}
//---------------------------------------------------------------------------
__fastcall TFarList::~TFarList()
{
  delete FListItems;
}
//---------------------------------------------------------------------------
void __fastcall TFarList::Put(int Index, const AnsiString S)
{
  if (DialogItem->Dialog->Handle)
  {
    FNoDialogUpdate = true;
    try
    {
      TStringList::Put(Index, S);
      FarListItem * ListItem = &FListItems->Items[Index];
      strcpy(ListItem->Text,
        Strings[Index].SubString(1, sizeof(ListItem->Text) - 1).c_str());

      FarListUpdate ListUpdate;
      memset(&ListUpdate, 0, sizeof(ListUpdate));
      ListUpdate.Index = Index;
      ListUpdate.Item = *ListItem;
      DialogItem->SendMessage(DM_LISTUPDATE, (int)&ListUpdate);
    }
    __finally
    {
      FNoDialogUpdate = false;
    }
  }
  else
  {
    TStringList::Put(Index, S);
  }
}
//---------------------------------------------------------------------------
void __fastcall TFarList::Changed()
{
  TStringList::Changed();

  if ((UpdateCount == 0) && !FNoDialogUpdate)
  {
    int PrevSelected;
    int PrevTopIndex;
    if (DialogItem->Dialog->Handle)
    {
      PrevSelected = Selected;
      PrevTopIndex = TopIndex;
    }
    if (FListItems->ItemsNumber != Count)
    {
      FListItems->ItemsNumber = Count;
      delete FListItems->Items;
      if (Count)
      {
        FListItems->Items = new FarListItem[Count];
      }
      else
      {
        FListItems->Items = NULL;
      }
    }
    for (int i = 0; i < Count; i++)
    {
      memset(&FListItems->Items[i], 0, sizeof(FListItems->Items[i]));
      strcpy(FListItems->Items[i].Text,
        Strings[i].SubString(1, sizeof(FListItems->Items[i].Text) - 1).c_str());
    }
    if (DialogItem->Dialog->Handle)
    {
      DialogItem->Dialog->LockChanges();
      try
      {
        DialogItem->SendMessage(DM_LISTSET, (int)FListItems);
        if (PrevTopIndex + DialogItem->Height > Count)
        {
          PrevTopIndex = Count > DialogItem->Height ? Count - DialogItem->Height : 0;
        }
        SetCurPos((PrevSelected >= Count) ? (Count - 1) : PrevSelected,
          PrevTopIndex);
      }
      __finally
      {
        DialogItem->Dialog->UnlockChanges();
      }
    }
  }
}
//---------------------------------------------------------------------------
void __fastcall TFarList::SetSelected(int value)
{
  if (Selected != value)
  {
    if (DialogItem->Dialog->Handle)
    {
      UpdatePosition(value);
    }
    else
    {
      DialogItem->Data = Strings[value];
    }
  }
}
//---------------------------------------------------------------------------
void __fastcall TFarList::UpdatePosition(int Position)
{
  if (Position >= 0)
  {
    SetCurPos(Position, -1);
  }
}
//---------------------------------------------------------------------------
void __fastcall TFarList::SetCurPos(int Position, int TopIndex)
{
  assert(DialogItem->Dialog->Handle);
  FarListPos ListPos;
  ListPos.SelectPos = Position;
  ListPos.TopPos = TopIndex;
  DialogItem->SendMessage(DM_LISTSETCURPOS, (int)&ListPos);
}
//---------------------------------------------------------------------------
void __fastcall TFarList::SetTopIndex(int value)
{
  if (value != TopIndex)
  {
    SetCurPos(-1, value);
  }
}
//---------------------------------------------------------------------------
int __fastcall TFarList::GetPosition()
{
  return DialogItem->SendMessage(DM_LISTGETCURPOS, NULL);
}
//---------------------------------------------------------------------------
int __fastcall TFarList::GetTopIndex()
{
  int Result;
  if (Count == 0)
  {
    Result = -1;
  }
  else
  {
    FarListPos ListPos;
    DialogItem->SendMessage(DM_LISTGETCURPOS, reinterpret_cast<int>(&ListPos));
    Result = ListPos.TopPos;
  }
  return Result;
}
//---------------------------------------------------------------------------
int __fastcall TFarList::GetSelected()
{
  int Result;
  if (Count == 0)
  {
    Result = -1;
  }
  else if (DialogItem->Dialog->Handle)
  {
    Result = GetPosition();
  }
  else
  {
    Result = IndexOf(DialogItem->DialogItem->Data);
  }

  if ((Result < 0) && (Count > 0))
  {
    Result = 0;
  }

  return Result;
}
//---------------------------------------------------------------------------
long __fastcall TFarList::ItemProc(int Msg, long Param)
{
  if (Msg == DN_LISTCHANGE)
  {
    if ((Param < 0) || ((Param == 0) && (Count == 0)))
    {
      DialogItem->UpdateData("");
    }
    else
    {
      assert(Param >= 0 && Param < Count);
      DialogItem->UpdateData(Strings[Param]);
    }
  }
  return false;
}
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
__fastcall TFarListBox::TFarListBox(TFarDialog * ADialog) :
  TFarDialogItem(ADialog, DI_LISTBOX)
{
  FList = new TFarList(this);
  DialogItem->ListItems = FList->ListItems;
  FAutoSelect = asFocus;
  FDenyClose = false;
}
//---------------------------------------------------------------------------
__fastcall TFarListBox::~TFarListBox()
{
  SAFE_DESTROY(FList);
}
//---------------------------------------------------------------------------
long __fastcall TFarListBox::ItemProc(int Msg, long Param)
{
  bool Result;
  if (FList->ItemProc(Msg, Param))
  {
    Result = true;
  }
  else
  {
    Result = TFarDialogItem::ItemProc(Msg, Param);
  }

  if ((Msg == DN_MOUSECLICK) && (Items->Count > 0))
  {
    FDenyClose = true;
  }

  return Result;
}
//---------------------------------------------------------------------------
void __fastcall TFarListBox::Init()
{
  Items->UpdatePosition(Items->Selected);
  UpdateMouseReaction();
}
//---------------------------------------------------------------------------
void __fastcall TFarListBox::SetAutoSelect(TFarListBoxAutoSelect value)
{
  if (AutoSelect != value)
  {
    FAutoSelect = value;
    if (Dialog->Handle)
    {
      UpdateMouseReaction();
    }
  }
}
//---------------------------------------------------------------------------
void __fastcall TFarListBox::UpdateMouseReaction()
{
  SendMessage(DM_LISTSETMOUSEREACTION, AutoSelect == asAlways);
}
//---------------------------------------------------------------------------
void __fastcall TFarListBox::SetItems(TStrings * value)
{
  FList->Assign(value);
}
//---------------------------------------------------------------------------
void __fastcall TFarListBox::SetList(TFarList * value)
{
  SetItems(value);
}
//---------------------------------------------------------------------------
bool __fastcall TFarListBox::CloseQuery()
{
  if (FDenyClose)
  {
    FDenyClose = false;
    return false;
  }
  else
  {
    return true;
  }
}
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
__fastcall TFarComboBox::TFarComboBox(TFarDialog * ADialog) :
  TFarDialogItem(ADialog, DI_COMBOBOX)
{
  FList = new TFarList(this);
  DialogItem->ListItems = FList->ListItems;
  AutoSelect = true;
}
//---------------------------------------------------------------------------
__fastcall TFarComboBox::~TFarComboBox()
{
  SAFE_DESTROY(FList);
}
//---------------------------------------------------------------------------
long __fastcall TFarComboBox::ItemProc(int Msg, long Param)
{
  if (Msg == DN_EDITCHANGE)
  {
    strcpy(DialogItem->Data, ((FarDialogItem *)Param)->Data);
  }
  
  if (FList->ItemProc(Msg, Param))
  {
    return true;
  }
  else
  {
    return TFarDialogItem::ItemProc(Msg, Param);
  }
}
//---------------------------------------------------------------------------
void __fastcall TFarComboBox::Init()
{
  Items->UpdatePosition(Items->Selected);
}

